#include <complex.h>
#include "cubic.h"

/*------------------------------------------------------------------------
   Improve accuracy of roots of cubic.
   In practice cubic() returns accurate roots,
   and this routine proves unnecessary.
*/
int cubic_improve(a, b, c, rt)
REAL a, b, c;
roots3 *rt;
{
    int ier = 0;

    switch (rt->type) {
    case ONE:
	ier += croot_improve(0, a, b, c, &rt->x0);
	ier += czoot_improve(0, a, b, c, &rt->xm, &rt->xp);
	break;
    case THREE:
	ier += croot_improve(0, a, b, c, &rt->x0);
	ier += croot_improve(0, a, b, c, &rt->xm);
	ier += croot_improve(0, a, b, c, &rt->xp);
	break;
    case UPPER_DOUBLE:
	ier += croot_improve(0, a, b, c, &rt->x0);
	ier += croot_improve(1, a, b, c, &rt->xm);
	rt->xp = rt->xm;
	break;
    case LOWER_DOUBLE:
	ier += croot_improve(0, a, b, c, &rt->xp);
	ier += croot_improve(1, a, b, c, &rt->xm);
	rt->x0 = rt->xm;
	break;
    case TRIPLE:
	ier += croot_improve(2, a, b, c, &rt->x0);
	rt->xp = rt->xm = rt->x0;
	break;
    }

    return(ier);
}

/*------------------------------------------------------------------------
   Newton improvement of real root of cubic.
*/
int croot_improve(nderiv, a, b, c, x)
int nderiv;
REAL a, b, c, *x;
{
    int it, its;
    int itmax = 3, itsmax = 15;
    REAL adx, dp, dx, p;

/*
    printf("nderiv x err_loose err_tight: %d %lg %lg %lg\n",
	nderiv, (double)*x, ERR_LOOSE, ERR_TIGHT);
*/

#ifndef NOIEEE
    /* root is not finite */
    if (!finite(*x)) return(0);
#endif

    its = it = 0;
    while (1) {
	its++;
	switch (nderiv) {
	case 0:
	    p = ((*x + a) * *x + b) * *x + c;
	    dp = (3. * *x + 2. * a) * *x + b;
	    break;
	case 1:
	    p = (3. * *x + 2. * a) * *x + b;
	    dp = 6. * *x + 2. * a;
	    break;
	case 2:
	    p = 6. * *x + 2. * a;
	    dp = 6.;
	}
	dx = p / dp;
        *x = *x - dx;
	adx = fabs(dx) / (1. + fabs(*x));
/*
	printf("adx dx x dp p: %lg %lg %lg %lg %lg\n",
	    (double)adx, (double)dx, (double)*x, (double)dp, (double)p);
*/
	if (adx <= ERR_TIGHT) break;
	if (adx <= ERR_LOOSE) {
	    it++;
	    if (it >= itmax) break;
	} else {
	    it = 0;
	}
	/* too many iterations */
	if (its >= itsmax) {
	    return(1);
	}
    }

    return(0);
}

/*------------------------------------------------------------------------
   Newton improvement of complex root of cubic.
*/
int czoot_improve(nderiv, a, b, c, xm, xp)
int nderiv;
REAL a, b, c, *xm, *xp;
{
    int it, its;
    int itmax = 3, itsmax = 15;
    REAL adz;
    REAL complex dp, dz, p, z;

/*
    printf("nderiv x err_loose err_tight: %d (%.16lg + I %.16lg) %lg %lg\n",
	nderiv, (double)*xm, (double)*xp, ERR_LOOSE, ERR_TIGHT);
*/

#ifndef NOIEEE
    /* NaN or Infinity */
    if (!finite(*xm) || !finite(*xp)) return(0);
#endif

    z = *xm + I * *xp;
    its = it = 0;
    while (1) {
	its++;
	switch (nderiv) {
	case 0:
	    p = ((z + a) * z + b) * z + c;
	    dp = (3. * z + 2. * a) * z + b;
	    break;
	case 1:
	    p = (3. * z + 2. * a) * z + b;
	    dp = 6. * z + 2. * a;
	    break;
	case 2:
	    p = 6. * z + 2. * a;
	    dp = 6.;
	}
	dz = p / dp;
        z = z - dz;
	adz = cabs(dz) / (1. + cabs(z));
/*
	printf("adz dz z dp p: %lg (%lg + I %lg) (%lg + I %lg) (%lg + I %lg) (%lg + I %lg)\n",
	    (double)adz, (double)creal(dz), (double)cimag(dz), (double)creal(z), (double)cimag(z), (double)creal(dp), (double)cimag(dp), (double)creal(p), (double)cimag(p));
*/
	if (adz <= ERR_TIGHT) break;
	if (adz <= ERR_LOOSE) {
	    it++;
	    if (it >= itmax) break;
	} else {
	    it = 0;
	}
	/* too many iterations */
	if (its >= itsmax) {
	    return(1);
	}
    }

    return(0);
}
