/*------------------------------------------------------------------------
   Andrew J. S. Hamilton 10 Mar 2001
   Andrew.Hamilton@Colorado.EDU
   http://casa.colorado.edu/~ajsh

   Elliptic integral of the 1st kind

           tu
      1   /            dt
      -   |  --------------------------
      2   /                         1/2
         tl  (t^3 + a t^2 + b t + c)

   Compile with -DDOUBLE to treat REAL as double.

   Built on the netlib/slatec rf routine by
   Carlson B. C., Notis, E. M. & Pexton, R. L.

1. The Legendre form F(phi, k) of the elliptic integral of the 1st kind:

#include "cubic.h"
int elintF(REAL phi, REAL m, *e)

   where m = k^2 is the modulus.

2. If you know the coefficients of the cubic polynomial, and don't care
   about what the roots are;
   or, if you already know the roots of the cubic;
   then:

#include "cubic.h"
int elint1(int l, REAL tl, REAL tu, REAL a, REAL b, REAL c, REAL *e)

   The l option determines how a, b, c are interpreted,
   as described in the comments to elint1().

3. If you know the coefficients of the cubic, and would like to know
   not only the integral but also the roots of the cubic, then:

#include "cubic.h"

    int ier;
    REAL a, b, c;
    REAL tl, tu, e;
    roots3 rt;

    ier = cubic(a, b, c, &rt);
    if (ier != 0) ier = elintr1(tl, tu, &rt, &e);

------------------------------------------------------------------------*/

#include "cubic.h"

/*------------------------------------------------------------------------
   Legendre form of elliptic integral of 1st kind.

		 phi
                /        d phi
   F(phi, k) =  |  ------------------
                /            2    1/2
	       0   (1 - m sin phi)

                     2 
                  sin phi
                1   /          dt
             =  -   |  ----------------------
                2   /                     1/2
                   0   [t(1 - t)(1 - m t)]

   where m = k^2 is the squared modulus.

   Input:  phi = angle in radians.
	   m = squared modulus k^2.  Must satisfy m <= 1.

   Return: 0 = ok;
	   1, 2, 3 per elintr1 (should not happen);
	   4 = modulus m > 0.

   Output: e = value of F(phi, k).

*/
int elintF(phi, m, e)
REAL phi, m, *e;
{
#define PI	3.141592653589793238462643383279502884197
    int ier;
    REAL iph, f, ph, tl, tu;
    roots3 rt;

    if (m < 0.) {		/* m < 0 */
	iph = floor(phi / PI + .5);
	ph = phi - iph * PI;
	tl = 0.;
	tu = -pow(sin(ph), 2);
	rt.type = THREE;
	rt.x0 = -1.;
	rt.xm = 0.;
	rt.xp = -1. / m;
	if (tu == 0.) {
	    *e = ph / 2.;
	} else {
	    ier = elintr1(tl, tu, &rt, &f);
	    if (ier != 0) return(ier);
	    f /= -sqrt(-m);
	    *e = (ph >= 0.)? f : -f;
	}
	if (iph != 0.) {
	    tl = 0.;
	    tu = -1.;
	    ier = elintr1(tl, tu, &rt, &f);
	    if (ier != 0) return(ier);
	    f /= -sqrt(-m);
	    *e += 2. * iph * f;
	}
    } else if (m == 0.) {	/* m = 0 */
	*e = phi;
    } else if (m < 1.) {	/* 0 < m < 1 */
	iph = floor(phi / PI + .5);
	ph = phi - iph * PI;
	tl = 0.;
	tu = pow(sin(ph), 2);
	rt.type = THREE;
	rt.x0 = 0.;
	rt.xm = 1.;
	rt.xp = 1. / m;
	if (tu == 0.) {
	    *e = ph / 2.;
	} else {
	    ier = elintr1(tl, tu, &rt, &f);
	    if (ier != 0) return(ier);
	    f /= sqrt(m);
	    *e = (ph >= 0.)? f : -f;
	}
	if (iph != 0.) {
	    tl = 0.;
	    tu = 1.;
	    ier = elintr1(tl, tu, &rt, &f);
	    if (ier != 0) return(ier);
	    f /= sqrt(m);
	    *e += 2. * iph * f;
	}
    } else if (m == 1.) {	/* m = 1 */
	if (fabs(phi) < PI / 2.) {
	    *e = ASINH(tan(phi));
	} else {
	    *e = phi / 0.;
	}
    } else {			/* m > 1 */
#ifdef DEBUG
	printf(" elintF: modulus %lf should be <= 1\n", m);
#endif
	return(4);
    }

    return(0);
}

/*------------------------------------------------------------------------
   Elliptic integral of the 1st kind.

   Input:  l = 0 means a b c are coefficients of cubic polynomial;
	     = 1 means a b c are real roots of cubic;
	     = 2 means a is real root and b +- i c are complex roots of cubic.
	   tl, tu = lower, upper limits of integral.
	   a, b, c = coefficients or roots of cubic.

   Return: 0 = ok;
	   1 = limits tl or tu out of valid range;
	   2 = rf function failed;
	   3 = unrecognized or NONE root type.

   Output: e = elliptic integral, as follows:

   If l = 0, then a b c are coefficients of cubic:
           tu
      1   /            dt
      -   |  --------------------------
      2   /                         1/2
         tl  (t^3 + a t^2 + b t + c)
   Note that
           tu                                    -tl
      1   /            dt                    1   /           dt
      -   |  ---------------------------  =  -   |  --------------------------
      2   /                          1/2     2   /                         1/2
         tl  (-t^3 + a t^2 + b t + c)          -tu  (t^3 + a t^2 - b t + c)

   If l = 1, then a b c are real roots of cubic:
           tu
      1   /            dt
      -   |  --------------------------
      2   /                         1/2
         tl  [(t - a)(t - b)(t - c)]
   Note that
           tu                                   -tl
      1   /            dt                   1   /            dt
      -   |  --------------------------  =  -   |  --------------------------
      2   /                         1/2     2   /                         1/2
         tl  [(a - t)(b - t)(c - t)]          -tu  [(t + a)(t + b)(t + c)]

   If l = 2, then a is real root and b +- i c are complex roots of cubic:
           tu
      1   /             dt
      -   |  -----------------------------
      2   /                            1/2
         tl  [(t - a)((t - b)^2 + c^2)]
   Note that
           tu                                  -tl
      1   /           dt                   1   /           dt
      -   |  -------------------------  =  -   |  -------------------------
      2   /                        1/2     2   /                        1/2
         tl  [(a-t)((t-b)^2 + c^2)]          -tu  [(t+a)((t+b)^2 + c^2)]
*/
int elint1(l, tl, tu, a, b, c, e)
int l;
REAL tl, tu, a, b, c, *e;
{
    int ier;
    roots3 rt;

#ifndef NOIEEE
    /* initialize integral to NaN */
    *e = 0. / 0.;
#endif

    switch (l) {

    /* a b c are coefficients of cubic */
    case 0:
	ier = cubic(a, b, c, &rt);
	if (ier != 0) return(ier);
	break;

    /* a b c are real roots of cubic */
    case 1:
	rt.type = THREE;
	if (a < b) {
	    if (b < c) {		/* a < b < c */
		rt.x0 = a;  rt.xm = b;  rt.xp = c;
	    } else if (b == c) {	/* a < b = c */
		rt.type = UPPER_DOUBLE;
		rt.x0 = a;  rt.xm = b;  rt.xp = c;
	    } else if (a < c) {		/* a < c < b */
		rt.x0 = a;  rt.xm = c;  rt.xp = b;
	    } else if (a == c) {	/* a = c < b */
		rt.type = LOWER_DOUBLE;
		rt.x0 = a;  rt.xm = c;  rt.xp = b;
	    } else if (a > c) {		/* c < a < b */
		rt.x0 = c;  rt.xm = a;  rt.xp = b;
	    } else {
		rt.type = NONE;
#ifdef DEBUG
		printf(" elint1: 3rd coeff %lf %lf %lf of cubic is NaN\n", a, b, c);
#endif
		return(3);
	    }
	} else if (a == b) {
	    if (b < c) {		/* a = b < c */
		rt.type = LOWER_DOUBLE;
		rt.x0 = a;  rt.xm = b;  rt.xp = c;
	    } else if (b == c) {	/* a = b = c */
		rt.type = TRIPLE;
		rt.x0 = a;  rt.xm = b;  rt.xp = c;
	    } else if (b > c) {		/* c < a = b */
		rt.type = UPPER_DOUBLE;
		rt.x0 = c;  rt.xm = a;  rt.xp = b;
	    } else {
		rt.type = NONE;
#ifdef DEBUG
		printf(" elint1: 3rd coeff %lf %lf %lf of cubic is NaN\n", a, b, c);
#endif
		return(3);
	    }
	} else if (a > b) {
	    if (b > c) {		/* c < b < a */
		rt.x0 = c;  rt.xm = b;  rt.xp = a;
	    } else if (b == c) {	/* b = c < a */
		rt.type = LOWER_DOUBLE;
		rt.x0 = b;  rt.xm = c;  rt.xp = a;
	    } else if (a > c) {		/* b < c < a */
		rt.x0 = b;  rt.xm = c;  rt.xp = a;
	    } else if (a == c) {	/* b < a = c */
		rt.type = UPPER_DOUBLE;
		rt.x0 = b;  rt.xm = a;  rt.xp = c;
	    } else if (a < c) {		/* b < a < c */
		rt.x0 = b;  rt.xm = a;  rt.xp = c;
	    } else {
		rt.type = NONE;
#ifdef DEBUG
		printf(" elint1: 3rd coeff %lf %lf %lf of cubic is NaN\n", a, b, c);
#endif
		return(3);
	    }
	} else {
	    rt.type = NONE;
#ifdef DEBUG
	    printf(" elint1: 1st or 2nd coeff %lf %lf %lf of cubic is NaN\n", a, b, c);
#endif
	    return(3);
	}
	break;

    /* a is real root and b +- i c are complex roots of cubic */
    case 2:
	rt.type = ONE;
	rt.x0 = a;
	rt.xm = b;
	rt.xp = c;
	break;

    /* l option unrecognized */
    default:
	rt.type = NONE;
#ifdef DEBUG
	printf(" elint1: option l = %d should be 0 to 2\n", l);
#endif
	return(3);
    }

    ier = elintr1(tl, tu, &rt, e);

    return(ier);
}

/*------------------------------------------------------------------------
   Elliptic integral of the 1st kind from roots of cubic.

   Input:  tl, tu = lower, upper limits of integral.
	   rt = roots of cubic.

   Return: 0 = ok;
	   1 = limits tl or tu out of valid range;
	   2 = rf function failed;
	   3 = unrecognized or NONE root type.
*/
int elintr1(tl, tu, rt, e)
REAL tl, tu, *e;
roots3 *rt;
{
    int i, ier;
    REAL d, f, fl, fu, p, q, s, sl, su, t, x, y, z;

#ifndef NOIEEE
    /* initialize integral to NaN */
    *e = 0. / 0.;
#endif

    /* one real root */
    if (rt->type == ONE) {
/*
          tu
      1   /               dt
      -   |  --------------------------------
      2   /                               1/2
         tl  [(t - x0)((t - xm)^2 + xp^2)]

   with x0 <= (tl & tu).
   The complete integral is from tl = x0 to tu = oo.
*/
	/* tl or tu below lower limit */
        if (tl < rt->x0 || tu < rt->x0) {
#ifdef DEBUG
	    printf(" elintr1: limits %lf %lf should be >= real root %lf\n", tl, tu, rt->x0);
#endif
	    return(1);
	}

	/* p = sqrt[(x0^2 - xm)^2 + xp^2] */
	q = rt->x0 - rt->xm;
	q = fabs(q);
	s = fabs(rt->xp);
	if (q == s) {
	    p = q * sqrt(2.);
	} else if (q > s) {
	    s = s / q;
	    p = q * sqrt(1. + s * s);
	} else {
	    q = q / s;
	    p = s * sqrt(1. + q * q);
	}

	/* do lower, upper limit in turn */
	for (i = 0; i < 2; i++) {
	    t = (i == 0)? tl : tu;
	    d = t - rt->x0;
	    if (d == 0.) {		/* t - x0 = 0 */
		f = 0.;
		s = -1.;
#ifndef NOIEEE
	    } else if (isinf(d)) {	/* t - x0 = oo */
		f = 0.;
		s = 1.;
#endif
	    } else if (d <= p) {	/* t - x0 <= p */
		q = 1. - d / p;
		x = q * q;
		q = 1. + d / p;
		z = q * q;
		q = (t - rt->xm) / p;
		s = rt->xp / p;
		y = q * q + s * s;
		ier = rf(x, y, z, &f);
		if (ier != 0) return(2);
		f *= sqrt(d) / p;
		s = -1.;
	    } else {			/* t - x0 > p */
		q = 1. - p / d;
		x = q * q;
		q = 1. + p / d;
		z = q * q;
		q = 1. + (rt->x0 - rt->xm) / d;
		s = rt->xp / d;
		y = q * q + s * s;
		ier = rf(x, y, z, &f);
		if (ier != 0) return(2);
		f /= sqrt(d);
		s = 1.;
	    }
	    if (i == 0) {
		fl = f;
		sl = s;
	    } else {
		fu = f;
		su = s;
	    }
	}
	*e = sl * fl - su * fu;
	if (sl * su < 0.) {
	    if (p == 0.) {
		*e = 1. / p;
	    } else {
		x = 0.;
		z = 1.;
		y = (1. + (rt->x0 - rt->xm) / p) / 2.;
		ier = rf(x, y, z, &f);
		if (ier != 0) return(2);
		f /= sqrt(p);
		*e += su * f;
	    }
	}

    /* three real roots */
    } else if (rt->type == THREE) {
	if (tl >= rt->x0 && tl <= rt->xm
	    && tu >= rt->x0 && tu <= rt->xm) {
/*
          tu
      1   /              dt
      -   |  -----------------------------
      2   /                            1/2
         tl  [(t - x0)(t - xm)(t - xp)]

   with x0 <= (tl & tu) <= xm.
   The complete integral is from tl = x0 to tu = xm.
*/
	    for (i = 0; i < 2; i++) {
		t = (i == 0)? tl : tu;
		d = t - rt->x0;
		if (d == 0.) {
		    f = 0.;
		} else {
		    p = rt->xm - rt->x0;
		    q = rt->xp - rt->x0;
		    x = (rt->xm - t) / d * q;
		    y = (rt->xp - t) / d * p;
		    z = q / d * p;
		    ier = rf(x, y, z, &f);
		    if (ier != 0) return(2);
		}
		if (i == 0) {
		    fl = f;
		} else {
		    fu = f;
		}
	    }
	    *e = fu - fl;

	} else if (tl >= rt->xp && tu >= rt->xp) {
/*
          tu
      1   /              dt
      -   |  -----------------------------
      2   /                            1/2
         tl  [(t - x0)(t - xm)(t - xp)]

   with xp <= (tl & tu).
   The complete integral is from tl = xp to tu = oo.
*/
	    for (i = 0; i < 2; i++) {
		t = (i == 0)? tl : tu;
		d = t - rt->x0;
#ifndef NOIEEE
		if (isinf(d)) {
		    f = 0.;
		} else {
#endif
		    x = (t - rt->xp) / d;
		    y = (t - rt->xm) / d;
		    z = 1.;
		    ier = rf(x, y, z, &f);
		    if (ier != 0) return(2);
		    f /= sqrt(d);
#ifndef NOIEEE
		}
#endif
		if (i == 0) {
		    fl = f;
		} else {
		    fu = f;
		}
	    }
	    *e = fl - fu;

	/* tl & tu are not both in [x0, xm] or both in [xp, oo] */
 	} else {
#ifdef DEBUG
	    printf(" elintr1: limits %lf %lf should be in [%lf, %lf], or >= %lf\n", tl, tu, rt->x0, rt->xm, rt->xp);
#endif
	    return(1);
	}

    /* upper root is degenerate double root */
    } else if (rt->type == UPPER_DOUBLE) {
/*
          tu
      1   /          dt
      -   |  -------------------
      2   /          1/2
         tl  (t - x0)   (t - xp)

   with x0 <= (tl & tu) <= xp, or xp <= (tl & tu).
   The integral converges at t -> x0, diverges at t -> xp, converges at t -> oo.
*/
	/* tl or tu below lower limit */
        if ((tl < rt->x0 || tu < rt->x0)
	/* tl and tu must both be same side of xp */
	    || (tl < rt->xp && tu > rt->xp)
	    || (tl > rt->xp && tu < rt->xp)) {
#ifdef DEBUG
	    printf(" elintr1: limits %lf %lf should be in [%lf, %lf] or >= %lf\n", tl, tu, rt->x0, rt->xp, rt->xp);
#endif
	    return(1);
	}

	p = rt->xp - rt->x0;
	q = sqrt(p);
#ifndef NOIEEE
	if (isinf(tl)) {
	    fu = sqrt(tu - rt->xp);
	    *e = -ASINH(q / fu) / q;
	} else if (isinf(tu)) {
	    fl = sqrt(tl - rt->xp);
	    *e = ASINH(q / fl) / q;
	} else {
#endif
	    d = tu - tl;
	    if (d == 0.) {
		*e = 0.;
	    } else {
		sl = sqrt(tl - rt->x0);
		su = sqrt(tu - rt->x0);
		fl = sqrt(fabs(tl - rt->xp));
		fu = sqrt(fabs(tu - rt->xp));
		*e = ASINH(q / fl / fu / (sl + su) * d) / q;
	    }
#ifndef NOIEEE
	}
#endif

    /* lower root is degenerate double root */
    } else if (rt->type == LOWER_DOUBLE) {
/*
          tu
      1   /          dt
      -   |  -------------------
      2   /                  1/2
         tl  (t - x0)(t - xp)

   with xp <= (tl & tu).
   The complete integral is from tl = xp to tu = oo.
*/
	/* tl or tu below lower limit */
        if (tl < rt->xp || tu < rt->xp) {
#ifdef DEBUG
	    printf(" elintr1: limits %lf %lf should be >= %lf\n", tl, tu, rt->xp);
#endif
	    return(1);
	}

	p = rt->xp - rt->x0;
	q = sqrt(p);
#ifndef NOIEEE
	if (isinf(tl)) {
	    su = sqrt(tu - rt->xp);
	    *e = -atan(q / su) / q;
	} else if (isinf(tu)) {
	    sl = sqrt(tl - rt->xp);
	    *e = atan(q / sl) / q;
	} else {
#endif
	    d = tu - tl;
	    if (d == 0.) {
		*e = 0.;
	    } else {
		sl = sqrt(tl - rt->xp);
		su = sqrt(tu - rt->xp);
		*e = atan(q / (p + sl * su) * d / (sl + su)) / q;
	    }
#ifndef NOIEEE
	}
#endif

    /* degenerate triple root */
    } else if (rt->type == TRIPLE) {
/*
          tu
      1   /     dt
      -   |  -----------
      2   /          3/2
         tl  (t - x0)

   with x0 <= (tl & tu).
   The integral diverges at t -> x0, converges at t -> oo.
*/
	/* tl or tu below lower limit */
        if (tl < rt->x0 || tu < rt->x0) {
#ifdef DEBUG
	    printf(" elintr1: limits %lf %lf should be >= %lf\n", tl, tu, rt->x0);
#endif
	    return(1);
	}

#ifndef NOIEEE
	if (isinf(tl)) {
	    su = sqrt(tu - rt->x0);
	    *e = -1. / su;
	} else if (isinf(tu)) {
	    sl = sqrt(tl - rt->x0);
	    *e = 1. / sl;
	} else {
#endif
	    d = tu - tl;
	    if (d == 0.) {
		*e = 0.;
	    } else {
		sl = sqrt(tl - rt->x0);
		su = sqrt(tu - rt->x0);
		*e = d / sl / su / (sl + su);
	    }
#ifndef NOIEEE
	}
#endif

    /* unrecognized or NONE root type */
    } else {
#ifdef DEBUG
	printf(" elintr1: root type %d is NONE or unknown\n", rt->type);
#endif
	return(3);
    }
    return(0);
}
