/*------------------------------------------------------------------------
   Elliptical integral of the 1st kind

           oo
      1   /             dt
      -   |  --------------------------
      2   /                         1/2
         0   [(t + x)(t + y)(t + z)]

   where x, y, and z are >=0 and at most one of them is 0.
   If one of x, y, z is zero, then the integral is complete.

   Compile with -DDOUBLE to treat REAL as double.

   Converted by Andrew J. S. Hamilton to c
   from netlib/slatec fortran routine rf.f, written by:

   Carlson, B. C.
     Ames Laboratory-DOE
     Iowa State University
     Ames, IA  50011
   Notis, E. M.
     Ames Laboratory-DOE
     Iowa State University
     Ames, IA  50011
   Pexton, R. L.
     Lawrence Livermore National Laboratory
     Livermore, CA  94550

   The original routine includes the following comments.
vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
   The duplication theorem is iterated until the variables are
   nearly equal, and the function is then expanded in Taylor
   series to fifth order.

   B. C. Carlson and E. M. Notis, Algorithms for incomplete
     elliptic integrals, ACM Transactions on Mathematical
     Software 7, 3 (September 1981), pp. 398-403.
   B. C. Carlson, Computing elliptic integrals by
     duplication, Numerische Mathematik 33, (1979), pp. 1-16.
   B. C. Carlson, Elliptic integrals of the first kind,
     SIAM Journal of Mathematical Analysis 8, (1977), pp. 231-242.

   ERRTOL - Relative error due to truncation is less than
            ERRTOL^6 / [4 * (1 - ERRTOL)]  .

   The accuracy of the computed approximation to the integral
   can be controlled by choosing the value of ERRTOL.
   Truncation of a Taylor series after terms of fifth order
   introduces an error less than the amount shown in the
   second column of the following table for each value of
   ERRTOL in the first column.  In addition to the truncation
   error there will be round-off error, but in practice the
   total error from both sources is usually less than the
   amount given in the table.

   Sample choices:  ERRTOL   Relative Truncation
                             error less than
                    1.0D-3    3.0D-19
                    3.0D-3    2.0D-16
                    1.0D-2    3.0D-13
                    3.0D-2    2.0D-10
                    1.0D-1    3.0D-7

   Decreasing ERRTOL by a factor of 10 yields six more
   decimal digits of accuracy at the expense of one or
   two more iterations of the duplication theorem.
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
------------------------------------------------------------------------*/

#include "cubic.h"

#ifdef DOUBLE
#  define ERRTOL	2.e-3
#else
#  define ERRTOL	3.e-2
#endif

int rf(x, y, z, f)
REAL x, y, z, *f;
{
    static REAL c0 = -1./10., c1 = 1./24., c2 = -3./44., c3 = 1./14.;
    REAL adxn, adyn, adzn, dxn, dyn, dzn, epsilon, e2, e3, lambda, mu, s,
	sxn, syn, szn, xn, yn, zn;

#ifndef NOIEEE
    /* x, y, z should >= 0 */
    if (x < 0. || y < 0. || z < 0.) {
	*f = 0. / 0.;
#ifdef DEBUG
	printf(" rf: x y z = %lf %lf %lf should be >= 0\n", x, y, z);
#endif
	return(1);	/* NaN */
    }
#endif

    xn = x;
    yn = y;
    zn = z;
    mu = (xn + yn + zn) / 3.;

#ifndef NOIEEE
    /* integral is NaN if any of x, y, or z is NaN */
    if (isnan(mu)) {
	*f = 0. / 0.;
#ifdef DEBUG
	printf(" rf: x y z = %lf %lf %lf contain NaN\n", x, y, z);
#endif
	return(0);
    }
#endif

#ifndef NOIEEE
    /* integral is 0 if x, y, z sum to oo */
    if (isinf(mu)) {
#ifdef DEBUG
	printf(" rf: x y z = %lf %lf %lf sum to oo\n", x, y, z);
#endif
	*f = 0.;
	return(0);
    }
#endif

#ifndef NOIEEE
    /* integral is oo if x, y, z are all 0 */
    if (mu == 0.) {
#ifdef DEBUG
	printf(" rf: x y z = %lf %lf %lf are all 0\n", x, y, z);
#endif
	*f = 1. / 0.;
	return(0);
    }
#endif

    while (1) {
	dxn = 1. - xn / mu;
	dyn = 1. - yn / mu;
	dzn = 1. - zn / mu;
	adxn = fabs(dxn);
	adyn = fabs(dyn);
	adzn = fabs(dzn);
	epsilon = (adxn > adyn)? adxn : adyn;
	epsilon = (epsilon > adzn)? epsilon : adzn;
	if (epsilon < ERRTOL) break;
	sxn = sqrt(xn);
	syn = sqrt(yn);
	szn = sqrt(zn);
	lambda = sxn * (syn + szn) + syn * szn;
	xn = (xn + lambda) / 4.;
	yn = (yn + lambda) / 4.;
	zn = (zn + lambda) / 4.;
	mu = (xn + yn + zn) / 3.;
    }
    e2 = dxn * dyn - dzn * dzn;
    e3 = dxn * dyn * dzn;
    s = c3 * e3 + (c2 * e3 + c1 * e2 + c0) * e2 + 1.;
    *f = s / sqrt(mu);

    return(0);		/* ok */
}
